/*!
 *  \file    png2c_ucg.c
 *  \author  Wim Dolman (<a href="email:w.e.dolman@hva.nl">w.e.dolman@hva.nl</a>)
 *  \date    22-02-2020
 *  \version 3.0
 *
 *  \brief   Conversion from PNG-file to C
 *
 *           This file uses functions from the library 'libpng' to read data
 *           from a PNG-file and convert the image to a C-format.
 *           It splits the picture in small blocks
 *
 *           For reading png see:
 *               http://zarb.org/~gc/html/libpng.html (Guillaume Cottenceau)
 *               https://gist.github.com/niw/5963798 (Yoshimasa Niwa)
 *               https://gist.github.com/jeroen/10eb17a9fb0e5799b772 (Jeroen Ooms)
 *
 *  build:   gcc -Wall -o png2c_ucg png2c_ucg.c -lpng
 *
 *           To install the 64-bit libpng with MinGW Msys2 use:
 *           pacman -S mingw64/mingw-w64-x86_64-libpng  
 *
 *  usage:   png2c_ucg png-file [-p|e|f] [-b <blocksize>] [-n <name>] [h-file]
 *           arguments:
 *                   png-file  the PNG file name is always the first argument 
 *
 *                         -p  use PROGMEM to indicate to use program space 
 *                         -e  use __memx  to indicate to use program space 
 *                         -f  use __flash to indicate to use program space (default)
 *
 *             -b <blocksize>  sets the block size  (default is 16384)
 *
 *                  -n <name>  sets name of block and inline function 
 *                             (default is basename png-file)
 *                     h-file  sets name of h-file (default uses basename png-file)
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#include <png.h>

#define VERSION             "3.0" 
#define NAME_SIZE           100 
#define DEFAULT_BLOCK_SIZE  16384 
#define DEFAULT_PREFIX      "_image_"

// data structure for images
typedef struct png_image {
  // info necessary for ucg blocks
  char      name[NAME_SIZE];   //!< image name
  char      uname[NAME_SIZE];  //!< image name uppercase 
  int       memoryType;        //!< memory type, 0: PROGMEM, 1: __flash, 2: __memx
  int       nPixelBytes;       //!< number of bytes of a pixel
  int       maxNlines;         //!< maximum number of lines in a block
  int       nBlocks;           //!< number of blocks   
  int       blockSize;         //!< block size
                               
  // info from PNG file
  int       width;             //!< number of pixels on a row of read image
  int       height;            //!< number rows of read image
  png_byte  color_type;        //!< color type of read image
  png_byte  bit_depth;         //!< bit depth of read image                         
  png_structp  png_ptr;        //!< pointer to internal struct of image
  png_infop    info_ptr;       //!< pointer to info struct of image
  png_bytep    *row_pointers;  //!< pointer to image array
} png_image_t;

// prototypes
void read_png_file(FILE *fp, png_image_t *image);
void clean_png(png_image_t *pimage);
void write_c_comment(FILE *fp, int argc, char *argv[]);
void write_c_ucg_blocks(FILE *fp, png_image_t *pimage);
void write_c_block(FILE *fp, png_image_t *pimage, 
                             int firstRow, int nRows, int iBlock);
void write_c_blocks(FILE *fp, png_image_t *pimage);
void write_c_ucg_image(FILE *fp, png_image_t *pimage);
void exit_with_message(const char *s, ...);


/*! \brief Main program
 *
 *  \param argc  number of aguments
 *  \param argv  the arguments:
 *               <ul><li>first argument png filename,
 *               <li>    second argument, is optional, c filename
 *               </ul>
 *  \details  this program read a pngfile and write the image
 *            as a c-format to the c-file or to stdout in
 *            case no c filename is given.
 *
 *  \return 0 if ok, 1 if not ok
 */

int main(int argc, char *argv[])
{
  FILE  *pngFile;
  FILE  *cFile;
  png_image_t  image;
  char   pngBasename[NAME_SIZE];
  char   cFilename[NAME_SIZE];
  int    i;
  char  *p;

  setbuf(stdout,NULL);
  if (argc < 2) {
    exit_with_message("Usage: %s <png-file> [-p] [-n <name>] [<file_out>]", argv[0]);
  }

  // get basename png-file
  if ( (p = strrchr(argv[1], '/')) == NULL ) {
    strcpy(pngBasename, argv[1]);
  } else {
    strcpy(pngBasename, p+1);
  }
  *(strrchr(pngBasename, '.')) = '\0';
  
  // read png file
  if ( (pngFile = fopen(argv[1], "rb")) == NULL ) {
    exit_with_message("File %s could not be opened for reading", argv[1]);
  }
  read_png_file(pngFile, &image);
  fclose(pngFile);

  // default options
  cFile = stdout;
  strcpy(image.name, ""); 
  strcpy(image.uname, ""); 
//  use_flash = 1;
  image.memoryType = 1;
  image.blockSize  = DEFAULT_BLOCK_SIZE;

  image.nPixelBytes = 1;
  if ( (image.color_type == PNG_COLOR_TYPE_RGB) || (image.color_type == PNG_COLOR_TYPE_RGBA) ) {
    image.nPixelBytes = 3;
  }  else {
    exit_with_message("Only color type RGB and RGBA are implemented, found %d\n", image.color_type); 
  }
  
  // get options
  i = 2;
  if (i<argc) {
    printf("Opties:\n");
  }
  while (i < argc) {
    if (argv[i][0] == '-') {
      switch (argv[i][1]) {
        case 'p' :   printf("\toptie p\n");  
                     image.memoryType = 0;
                     break; 
        case 'e' :   printf("\toptie e\n");  
                     image.memoryType = 2;
                     break; 
        case 'f' :   printf("\toptie f\n");  
                     image.memoryType = 1;
                     break; 
        case 'b' :   printf("\toptie b\n");  
                     i++;
                     image.blockSize = atoi(argv[i]);
                     if (image.blockSize < image.width*image.nPixelBytes) {
                       image.blockSize = image.width*image.nPixelBytes;
                       printf("\tblocksize (changes to minimal value) %d\n", image.blockSize);
                     } else{
                       printf("\tblocksize %d\n", image.blockSize);
                     }
                     break; 
        case 'n' :   printf("\toptie n\n");  
                     i++;
                     printf("\tnaam %s\n", argv[i]);
                     strncpy(image.name, argv[i], NAME_SIZE-1);
                     image.name[NAME_SIZE-1] = '\0';                     
                     break; 
      }
    } else {
      printf("\th-file: %s\n", argv[i]);
      if ( (cFile = fopen(argv[i], "wb")) == NULL ) {
        clean_png(&image);
        exit_with_message("File %s could not be opened for writing", argv[i]);
      }      
    }
    i++;
  }
  
  // get name an upper case name
  if ( strcmp(image.name, "") == 0 ) {
    strncpy(image.name, pngBasename, NAME_SIZE-1);
    image.name[NAME_SIZE-1] = '\0';
  }
  strncpy(image.uname, image.name, NAME_SIZE-1);   
  strupr(image.uname);   
  
  // calculate maximum number lines of a block
  image.maxNlines = image.blockSize/(image.width*image.nPixelBytes);
  if (image.maxNlines > image.height) image.maxNlines = image.height;
  image.nBlocks   = image.height/image.maxNlines;
  if (image.maxNlines*image.nBlocks < image.height) image.nBlocks++;

  // open c file 
  if (cFile == stdout) {
    strncpy(cFilename, image.name, NAME_SIZE-3);
    image.name[NAME_SIZE-3] = '\0';
    strncat(cFilename, ".h", 3);
    image.name[NAME_SIZE-1] = '\0';
    if ( (cFile = fopen(cFilename, "wb")) == NULL ) {
      clean_png(&image);
      exit_with_message("File %s could not be opened for writing", cFilename);
    }      
  }
    
  // write c file
  write_c_comment(cFile, argc, argv);
  write_c_ucg_blocks(cFile, &image);
  fclose(cFile);

  // clean up allocated memory
  clean_png(&image);
  png_destroy_read_struct(&image.png_ptr, &image.info_ptr, NULL);

  return 0;
}

/*! \brief Read PNG file
 *
 *  \param  fp      file pointer to PNG-file
 *  \param  image   file pointer to image
 *
 *  \return void
 */
void read_png_file(FILE *fp, png_image_t *pimage)
{
  unsigned char header[8];  // 8 is the maximum size that can be checked

  // test if a png-file
  fread(header, 1, 8, fp);
  if ( png_sig_cmp(header, 0, 8) ) {
    fclose(fp);
    exit_with_message("[read_png_file] File is not recognized as a PNG file");
  }

  // creates structs for png
  if ( (pimage->png_ptr=png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL)) == NULL ) {
    fclose(fp);
    exit_with_message("[read_png_file] png_create_read_struct failed");
  }
  if ( (pimage->info_ptr= png_create_info_struct(pimage->png_ptr)) == NULL) {
    fclose(fp);
    exit_with_message("[read_png_file] png_create_info_struct failed");
  }

  // setjmp for error in initialisation
  if (setjmp(png_jmpbuf(pimage->png_ptr))) {
    png_destroy_read_struct(&pimage->png_ptr, &pimage->info_ptr, NULL);
    fclose(fp);
    exit_with_message("[read_png_file] Error during init_io");
  }

  png_init_io(pimage->png_ptr, fp);
  png_set_sig_bytes(pimage->png_ptr, 8);
  png_read_info(pimage->png_ptr, pimage->info_ptr);

  pimage->width      = png_get_image_width(pimage->png_ptr, pimage->info_ptr);
  pimage->height     = png_get_image_height(pimage->png_ptr, pimage->info_ptr);
  pimage->color_type = png_get_color_type(pimage->png_ptr, pimage->info_ptr);
  pimage->bit_depth  = png_get_bit_depth(pimage->png_ptr, pimage->info_ptr);

  //number_of_passes = png_set_interlace_handling(png_ptr);
  png_read_update_info(pimage->png_ptr, pimage->info_ptr);

  // setjmp for error in reading
  if (setjmp(png_jmpbuf(pimage->png_ptr))) {
    png_destroy_read_struct(&pimage->png_ptr, &pimage->info_ptr, NULL);
    fclose(fp);
    exit_with_message("[read_png_file] Error during read_image");
  }

  // allocate memory
  pimage->row_pointers = (png_bytep*) malloc(sizeof(png_bytep) * pimage->height);
  for (int y=0; y<pimage->height; y++)
    pimage->row_pointers[y] = (png_byte*) malloc(png_get_rowbytes(pimage->png_ptr, pimage->info_ptr));

  // read image
  png_read_image(pimage->png_ptr, pimage->row_pointers);

  fclose(fp);
}

/*! \brief Clear png date
 *
 *  \param row_pointers   pointer to array with pointer pointing to rows
 *
 *  \return void
 */
void clean_png(png_image_t *pImage)
{
  int y;

  for (y=0; y<pImage->height; y++) {
    free(pImage->row_pointers[y]);
  }

  free(pImage->row_pointers);
}

/*! \brief Write c file of the image in ucg blocks
 *
 *  \param  fp          file pointer to c-file (h-file)
 *  \param  pimage      pointer to image
 *
 *  \return void
 */
void write_c_ucg_blocks(FILE *fp, png_image_t *pimage)
{
  fprintf(fp, "#ifndef _IMAGE_%s_H_\n#define __%s_H_ 1\n\n", pimage->uname, pimage->uname);
  write_c_blocks(fp, pimage);
  write_c_ucg_image(fp, pimage);
  fprintf(fp, "#endif // _IMAGE_%s_H_", strupr(pimage->uname));
}

/*! \brief  Write a image block
 *
 *  \param  fp          file pointer to c-file (h-file)
 *  \param  pimage      pointer to image
 *  \param  firstRow    number of the first row
 *  \param  nRows       number of rows in the image block
 *  \param  iBlock      index of the image block
 *
 *  \return void
 */
void write_c_block(FILE *fp, png_image_t *pImage, int firstRow, int nRows, int iBlock)
{
  png_bytep px;

  if ( pImage->memoryType == 2 ) {
    fprintf(fp, "static const __memx uint8_t %s%s_%d[] ={\n  ", DEFAULT_PREFIX, pImage->name, iBlock);
  } else if ( pImage->memoryType == 1 ) {
    fprintf(fp, "static const __flash uint8_t %s%s_%d[] ={\n  ", DEFAULT_PREFIX, pImage->name, iBlock);
  } else {
    fprintf(fp, "static const uint8_t %s%s_%d[] PROGMEM = {\n  ", DEFAULT_PREFIX, pImage->name, iBlock);
  }

  for(int y = firstRow; y < firstRow+nRows; y++) {

    png_bytep row = pImage->row_pointers[y];    
    for(int x = 0; x < pImage->width; x++) {
      if (pImage->color_type == PNG_COLOR_TYPE_RGBA) {
        px = &(row[x * 4]); 
      } else {
        px = &(row[x * 3]); 
      }        
      fprintf(fp, "0x%02X, 0x%02X, 0x%02X", px[0], px[1], px[2]); // with RGBA alpha is skipped
      if (x < pImage->width-1) {
        fprintf(fp, ", ");
      } else {
        if (y == firstRow+nRows-1) {
        fprintf(fp, "\n};\n\n");
        } else {
          fprintf(fp, ",\n  ");
        }
      }
    } // end printing row of pixels (x) 
    
  } // end printing images (y)
}

/*! \brief  Write an image in blocks
 *
 *  \param  fp          file pointer to c-file (h-file)
 *  \param  pimage      pointer to image
 *
 *  \return void
 */
void write_c_blocks(FILE *fp, png_image_t *pImage)
{
  int firstRow;
  int currNRows;

  firstRow = 0;
  currNRows  = pImage->maxNlines;
  for(int i=0; i<pImage->nBlocks; i++) {
    write_c_block(fp, pImage,  firstRow, currNRows, i);
    firstRow = firstRow + currNRows;
    if ( (firstRow + currNRows) > pImage->height ) {
      currNRows = pImage->height - firstRow;
    } 
  }
}

/*! \brief  Write the inline function that calls the ucg print function
 *
 *  \param  fp          file pointer to c-file (h-file)
 *  \param  pimage      pointer to image
 *
 *  \return void
 */
void write_c_ucg_image(FILE *fp, png_image_t *pImage)
{
  int firstRow;
  int currNRows;
  
  fprintf(fp, "static inline void ucg_image_%s(ucg_t *ucg, ucg_int_t x, ucg_int_t y)\n{\n", 
              pImage->name);
  firstRow = 0;
  currNRows  = pImage->maxNlines;
  for(int i = 0; i<pImage->nBlocks; i++) {
    if ((firstRow + pImage->maxNlines) > pImage->height) {
      currNRows = pImage->height - firstRow;
    }
    fprintf(fp, "  ucg_BitmapPrint(ucg, x, y + %3d, %3d, %3d, %d, %s%s_%d);\n", 
                   firstRow, pImage->width, currNRows,
                   pImage->nPixelBytes, DEFAULT_PREFIX, pImage->name, i);
    firstRow = firstRow + currNRows;
  }
  fprintf(fp, "}\n\n");
}

/*! \brief Write comment to c file
 *
 *  \param  fp          file pointer to c-file (h-file)
 *  \param  argc        number of arguments (is argc from main) 
 *  \param  argv[]      pointer to arguments (is argv[] from main) 
 *
 *  \return void
 */

void write_c_comment(FILE *fp, int argc, char *argv[])
{
  time_t t = time(NULL);
  struct tm tm = *localtime(&t);
  fprintf(fp, "//\n// This image file is generated on %d-%02d-%02d with png2c_ucg version %s:\n", 
              tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, VERSION);
  fprintf(fp, "//     png2c_ucg %s", argv[1]);
  for(int i=2; i<argc; i++) {
    fprintf(fp, " %s", argv[i]);
  }
  fprintf(fp, "\n//\n");
}


/*! \brief Exit with message
 *
 *  \param s   pointer to (format)string
 *
 *  \details   First this function print a (formatted) messagestring
 *             to stderr.
 *             After that it exit the program
 *
 *  \return void
 */
void exit_with_message(const char *s, ...)
{
  va_list vl;

  va_start(vl, s);
  vfprintf(stderr, s, vl);
  fprintf(stderr, "\n");
  va_end(vl);

  exit(1);
}


